*                       Array BINARY SEARCH & FIND
*                          PWITTE May 1st 1994
*
        section code

        xdef    BINSEARCH
        xdef    ms_start
        xdef    ms_ret

        xref    indinfo             ;get parameters
        xref    str_cmp
        xref    int_cmp
        xref    flt_cmp
        xref    find
*
BINSEARCH
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* SuperBASIC function to locate an element in an array using a fast binary
* search algorithm. The array must have been previously sorted, and the search
* must use the same comparison criteria as was used in the sort. The position
* of the first matching element is returned or a negative position number
* indicating where the missing element would have been, -1. (-posn -1)
*
* psn% = BINSEARCH ( [index% ! ] [#]array [; [+|-]comparison_type% ] ,item )
*       where index% is an integer array of no more than DIMN(array) elements
*             each element being a unique pointer to an element in array
*       and   array is a single or multiple (but does it make sense?)
*             dimensioned array of any current qdos type, or an array slice
*             of more than one element
*        or   a reference to an array as above preceded by a hash
*       and   comparison_type% is the type of comparison required
*             types are as for +/-(Qdos type + 1)
*             Default = 1 = Qdos type 0
*             The comparison type sign indicates ascending ( + implicit and
*             default) or descending ( - ) order
*       and   item is the item being searched for. This will be treated as
*             of the same type as the array.
*
        moveq   #0,d1               ;flag SEARCH
ms_start
        bsr     indinfo             ;get index, array & cmp info
        bne.s   ms_rts
ms_entry
        move.l  $58(a6),a1
        movem.l d1/d6/a1/a2/a4,-(a7)
        tst.w   d3                  ;this is type-specific info
        beq.s   ms_flt
        bmi.s   ms_str
ms_int
        move.w  $112,a2             ;ca.gtint
        lea.l   int_cmp,a4
        bra.s   ms_cont
ms_rts
        rts
ms_str
        move.w  $116,a2             ;ca.gtstr
        lea.l   str_cmp,a4
        bra.s   ms_cont
ms_flt
        move.w  $114,a2             ;ca.gtfp
        lea.l   flt_cmp,a4
ms_cont
        jsr     (a2)                ;get relevant parameter
        movem.l (a7)+,d1/d6/a0/a2/a5  ;a0->stack base a2->index, a5->array
        bne.s   ms_rts
        move.l  a0,-(a7)            ;store stack base pointer
        tst.l   d1                  ;FIND or SEARCH?
        bmi     find

        subq.l  #1,d7               ;convert count to ptr to last element
        move.l  a2,d0               ;indexed?
        beq.s   ms_dir              ; no
* ms_ind
        asl.l   #1,d7               ;convert to word steps
        add.l   a2,d7               ;add index base addr
        move.l  d7,a3               ;a3 -> last a2 -> first
        move.l  a2,-(a7)            ;save index base address
        bsr.s   ind_search
        sub.l   (a7)+,d4            ;take away index base address
        asr.l   #1,d4
        tst.l   d0                  ;found or not?
        bra.s   ms_return
ms_dir
        move.l  d7,a3               ;a3 -> last a2->first (a2=0)
        bsr.s   dir_search
ms_return
        beq.s   ms_ret
        neg.w   d4
        subq.w  #1,d4
ms_ret
        move.l  (a7)+,a1            ;restore stack base pointer
        subq.l  #2,a1               ;room for integer result
        move.l  a1,$58(a6)
        move.w  d4,0(a6,a1.l)       ;put result in place
        moveq   #3,d4
        moveq   #0,d0
        rts
*
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*                      Binary search sub-routine (direct)
*
* input : d5.l  comp type (see marrinfo for details)
*         d6.l = element size
*         a0 =  running element pointer
*         a1 -> comparison variable
*         a2 =  lo
*         a3 =  hi
*         a4 -> comp routine
*         a5 =  array base address
* output : d0 = result of search: 0 = found, not 0 = not found
*         d4 -> insertion position ("found" or "not found" at)
*         a5    unchanged
*         flags set
*         d0-d4 smashed.
*
* Notes:
* From dir_sex the point has been to adjust the result so as to hit the *first*
* position from the start of the element list that matches the criteria. A pure
* binary search may end on a random target if there are multiple matches:
* that does not produce a stable sort
*
dir_nf
        move.l  d2,d4
        moveq   #-1,d0
        rts
dir_search
        move.l  a3,d3
        move.l  a2,d2               ;a2 -> lo
dir_sloop
        cmp.l   d3,d2               ;hi<lo?
        bgt.s   dir_nf              ; yes, exit "not found"
        move.l  d2,d4
        add.l   d3,d4
        asr.l   #1,d4               ;(hi+lo) div 2

        move.l  d4,d1
        mulu    d6,d1               ;convert to address offset
        lea.l   0(a5,d1.l),a0       ;add in base address
        jsr     (a4)
        beq.s   dir_sex
        bgt.s   dir_slo
* dir_shi
        move.l  d4,d2
        addq.l  #1,d2
        bra.s   dir_sloop
dir_slo
        move.l  d4,d3
        subq.l  #1,d3
        bra.s   dir_sloop
dir_sex
        cmp.l   d4,d2               ;are we at elno 0?
        bpl.s   dir_sret            ; yea, quit (should it be beq.s ?
        subq.l  #1,d4               ; no, try prior element
        move.l  d4,d1
        mulu    d6,d1
        lea.l   0(a5,d1.l),a0
        jsr     (a4)
        beq.s   dir_sex
        addq.l  #1,d4               ;adjust overshoot,
        moveq   #0,d0               ; signal "found" and
dir_sret
        rts                         ; return
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*                      Binary search sub-routine (indexed)
*
* input : d5.l  comp type (see marrinfo for details)
*         d6.l = element size
*         a0 =  running element pointer
*         a1 -> comparison variable
*         a2 ->> lo
*         a3 ->> hi (both include index base address)
*         a4 -> comp routine
*         a5 =  array base address
* output : d0 = result of search: 0 = found, not 0 = not found
*         d4 ->> insertion position ("found" or "not found" at)
*         a5    unchanged
*         flags set
*         d0-d4 smashed.
*
* Notes:
* From ind_sex the point has been to adjust the result so as to hit the *first*
* position from the start of the element list that matches the criteria. A pure
* binary search may end on a random target if there are multiple matches:
* that does not produce a stable sort
*
ind_nf
        move.l  d2,d4
        moveq   #-1,d0
        rts
ind_search
        move.l  a3,d3
        move.l  a2,d2               ;a2 -> lo
ind_sloop
        cmp.l   d3,d2               ;hi<lo?
        bgt.s   ind_nf              ; yes, exit "not found"
        move.l  d2,d4
        add.l   d3,d4
        asr.l   #1,d4               ;(hi+lo) div 2

        bclr    #0,d4               ;round to nearest step size
        move.w  0(a6,d4.l),d1
        mulu    d6,d1
        add.l   a5,d1
        movea.l d1,a0
        jsr     (a4)
        beq.s   ind_sex
        bgt.s   ind_slo
* ind_shi
        move.l  d4,d2
        addq.l  #2,d2
        bra.s   ind_sloop
ind_slo
        move.l  d4,d3
        subq.l  #2,d3
        bra.s   ind_sloop
ind_sex
        cmp.l   d4,d2               ;are we at elno 0?
        bpl.s   ind_sret            ; yea, quit
        subq.l  #2,d4               ; no, try prior element
        move.w  0(a6,d4.l),d1
        mulu    d6,d1
        add.l   a5,d1
        movea.l d1,a0
        jsr     (a4)
        beq.s   ind_sex
        addq.l  #2,d4               ;adjust overshoot,
        moveq   #0,d0               ; signal "found" and
ind_sret
        rts                         ; return
*
        end
